"""This module encapsulates and formalizes all aspects
and rules of the Online Packing Problem.

For each instant, the algorithm receives k options of items, each one with a value and a cost-vector.
The algorithm chooses which option to pack, and then waits until receiving the options for
the next instant. The algorithm can also choose to pack no items at a given instant. Also, it may not
pack any item that that causes a violation of the capacity constraint in some dimension.
"""
from typing import List, Union
import copy


class PackingProblem:
    """Class that enforces all aspects of the
    online packing problem.

    This class is statefull, meaning it carries the current state
    of the game, and some methods depend on that state.

    Parameters
    ----------
    capacity : float or int
        The capacity for the problem (same in every dimension).
    cost_dimension : int
        The dimension of the cost vector.

    Methods
    -------
    set_current_inputs(values, costs)
        Set items' values and costs for the current instant.
    item_fits(idx)
        Checks if current instant's item of index `idx` fits in capacity constraints.
    pack(idx)
        Pack current instant's item of index `idx`.
    end_packing()
        Ends the packing phase.
    get_capacity()
        Get capacity of the problem instance.
    get_options_per_instant()
        Get the number of available items in each instant.
    get_available_values()
        Get the value of the items that are available for the current instant.
    get_available_costs()
        Get the cost of the items that are available for the current instant.
    get_cost_dimension()
        Get the dimension of the cost vectors.
    """

    _capacity: Union[int, float]
    _packed_items: List[int]
    _packed_value_sum: Union[int, float]
    _packed_costs_sum: List[Union[int, float]]
    _cost_dimension: int
    _options_per_instant: int
    # inputs seen in until current time
    # values: 1st index = day   //   2nd index = item value
    _available_values: List[List[Union[int, float]]]
    # costs: 1st index = day   //   2nd index = item costs vector   //   3rd index = item cost dimension
    _available_costs: List[List[List[Union[int, float]]]]

    def __init__(self, capacity: Union[float, int], cost_dimension: int):
        self._packed_value_sum = 0.0
        self._capacity = -1
        self._set_capacity(capacity)
        self._set_cost_dimension(cost_dimension)
        self._available_values = list()
        self._available_costs = list()
        self._packed_items = list()

    def _set_cost_dimension(self, cost_dimension: int) -> None:
        """Validates cost dimension and sets cost dimension.

        Parameters
        ----------
        cost_dimension : int
            The dimension of the cost vectors.

        Raises
        ------
        Exception
            If `cost_dimension` is lower than 1.
        """
        if cost_dimension <= 0:
            raise Exception("cost_dimension has to be greater than 0")
        self._cost_dimension = cost_dimension

    def _set_capacity(self, capacity: Union[float, int]):
        """Validates and sets capacity for the problem.

        Parameters
        ----------
        capacity : float or int
            The capacity of each dimension (they must be the same).

        Raises
        -------
        Exception
            If capacity already set.
        Exception
            If capacity is negative.
        """
        if capacity < 0 - 1e-6:
            raise Exception(f"Capacity must be >= 0 (tried setting to {capacity}).")
        elif self._capacity >= 0-1e-6:
            raise Exception(f"Capacity already set (capacity={self._capacity}).")
        else:
            self._capacity = capacity

    def get_capacity(self) -> Union[int, float]:
        """Get the problem capacity.

        Returns
        -------
        int or float
            Previously set problem capacity.
        """
        return self._capacity

    def get_options_per_instant(self) -> int:
        """Get the number of available items per instant.

        Returns
        -------
        int
            Number of available items per instant.
        """
        return self._options_per_instant

    def _validate_curr_inputs(self,
                              values: List[Union[int, float]],
                              costs: List[List[Union[int, float]]]) -> None:
        if len(values) != len(costs):
            raise Exception("len(values) != len(costs)")
        # validate values between 0 and 1:
        for v in values:
            if v < 0.0 - 1e-6 or v > 1 + 1e-6:
                raise Exception(
                    "Option values must be in range [0, 1].")
        for option in costs:
            if len(option) != self._cost_dimension:
                raise Exception("Received a cost vector with different dimension from previously seen.")
            for cost in option:
                if cost < 0 - 1e-6 or cost > 1 + 1e-6:
                    raise Exception("All option costs must in range [0, 1].")
        # validate number of options and cost-dimensions match previously seen:
        if len(self._available_values) > 0:
            if len(values) != self._options_per_instant:
                raise Exception(
                    "Number of available options in values in different from previously seen options.")
            if len(costs) != self._options_per_instant:
                raise Exception(
                    "Number of available options in costs is different from previously seen options.")

    def set_current_inputs(self, values: List[Union[int, float]],
                           costs: List[List[Union[int, float]]]) -> None:
        if len(self._packed_items) < len(self._available_values):
            raise Exception(
                "Tried to set next instant without choosing an option for the current instant.")
        self._validate_curr_inputs(values, costs)
        self._available_values.append(values)
        self._available_costs.append(costs)
        if len(self._available_values) == 1:
            self._options_per_instant = len(values)
            self._packed_costs_sum = [0.0 for _ in range(self._cost_dimension)]

    def item_fits(self, idx: int) -> bool:
        """Checks that currently available item of index `idx` fits the capacity constriants.

        Parameters
        ----------
        idx : int
            Index of the item to be verified.

        Returns
        -------
        bool
            True if the item fits the constraints, False otherwise.
        """
        if idx == -1:
            return True
        for dim in range(self._cost_dimension):
            if self._available_costs[-1][idx][dim] + self._packed_costs_sum[dim] > self._capacity + 1e-6:
                return False
        return True

    def pack(self, idx: int) -> Union[int, float]:
        """Pack item of index `idx`, from the currently available.

        Parameters
        ----------
        idx : int
            Index of the item to pack.

        Returns
        -------
        int or float
            Total value of the packed items so far.

        Raises
        ------
        Exception
            An item was already packed for the current instant.
        Exception
            Index for the item out of bounds.
        Exception
            Chosen item  exceed capacity in some dimension.
        """
        if len(self._available_values) == len(self._packed_items):
            raise Exception(
                "Packing already made. Currently waiting for the next input.")
        elif idx < -1 or idx > self._options_per_instant-1:
            raise Exception(f"Tried to pack item out of bounds. Available indexes to pack are [0, ..., \
                {self._options_per_instant-1}] or -1 to pack no items.")

        # pack chosen item
        if idx == -1:
            self._packed_items.append(idx)
            return self._packed_value_sum
        elif self.item_fits(idx):
            self._packed_items.append(idx)
            for dim in range(self._cost_dimension):
                self._packed_costs_sum[dim] += self._available_costs[-1][idx][dim]
            self._packed_value_sum += self._available_values[-1][idx]
            return self._packed_value_sum
        else:
            raise Exception("Tried to pack an item that exceeds capacity in some dimension")

    def get_available_values(self) -> List[List[Union[int, float]]]:
        """Get values of all the available items for past and current instants.

        Returns
        -------
        list of list of (int or float)
            List contianing, for each instant, a list with the values of the available items for \
            that instant.
        """
        return copy.deepcopy(self._available_values)

    def get_available_costs(self) -> List[List[List[Union[int, float]]]]:
        """Get costs of all the available items for past and current instants.

        Returns
        -------
        list of list of list of (int or float)
            List contianing, for each instant, a list with the cost vectors of the available items for \
            that instant.
        """
        return copy.deepcopy(self._available_costs)

    def get_cost_dimension(self):
        """Get the dimension of the cost vectors.

        Returns
        -------
        int
            The dimension of the cost vectors.
        """
        return self._cost_dimension



from abc import ABC, abstractmethod
from typing import Union, List


class BaseSolver(ABC):
    """Abstract class unifying methods for the offline solvers.

    Parameters
    ----------
    values : list of list of (int or float)
        A list of instants, where each instant contains is a list of the available values of that instant.
    costs : list of list of list of (int or float)
        A list of instants, where each instant is a list of options available, and each option
        is a list containing this option's costs for each dimension.
    capacity : float
        The capacity of each dimension (each dimension have the same capacity).

    Attributes
    ----------
    optimum_value : float or int
        The optimum value found for the LP problem.
    packed_items : list of int
        A list containing the indexes of the items chosen in each instant.
    packed_weight_sum : list of float
        A list containing the optimal solution's total cost for each dimension.

    Methods
    -------
    solve()
        Abstract method, should solve the LP problem.
    print_result()
        Prints a report containing the optimum solution and the total weights in each dimension.
    """
    _size: int
    _values: List[List[Union[float, int]]]
    _costs: List[List[List[Union[float, int]]]]
    _options_per_instant: int
    _cost_dimension: int
    optimum_value: Union[float, int]
    packed_items: List[int]
    packed_weight_sum: List[float]

    def __init__(self, values: List[List[Union[float, int]]],
                 costs: List[List[List[Union[float, int]]]],
                 capacity: float):
        self._size = len(values)
        self._options_per_instant = len(values[0]) if len(values) else 0
        self._cost_dimension = len(costs[0][0]) if len(costs) and len(costs[0]) else 0
        self._values = values
        self._costs = costs
        self._capacity = capacity
        self.optimum_value = 0.0
        self.packed_items = list()
        self.packed_weight_sum = [0.0 for _ in range(self._cost_dimension)]

    @abstractmethod
    def solve(self) -> None:
        """Abstract method that should solve the LP problem and set
        `optimum_value`, `packed_items` and `packed_weight_sum` attributes.
        """
        pass

    def print_result(self) -> None:
        """Prints a report containing the optimum solution
        and the total weights in each dimension
        """
        print(f"Total value = {self.optimum_value:.4f}")
        for dim in range(self._cost_dimension):
            print(
                f"\tTotal weight dim {dim}: {self.packed_weight_sum[dim]:.5f} / {self._capacity:.5f}")



"""Implements a MWU-max method.

    The MWU implemented here tries to maximize a cost function (gains function, in this case).
    This is done by alocating probabilities in the available dimensions, with the constraint that
    the sum of allocated probabilities is <= 1.
"""
from typing import List
import copy


class MwuMax:
    """Class that implements MWU-max algorithm.

    Parameters
    ----------
    n_experts : int
        Dimension of the problem cost function.
    eps : float
        Calibration parameter that regulates the step-size of the updates.

    Methods
    -------
    update_weights(expert_gains: List[float])
        Update mwu probabilities based on a new input of the cost function.
    get_probs()
        Get a copy of the currently allocated probabilities.
    get_n_experts()
        Get the number of dimensions used in this instance.

    Notes
    -----
    Usually, the mwu respects the following restriction: :math:`\\sum_{i=1}^d p_i = 1`, where :math:`p_i` is
    the probability or weight attributed to dimnesion :math:`i`. In order to accomodate the restriction that
    :math:`\\sum_{i=1}^d p_i \\leq 1`, we consider an extra dimension, which always receives cost 0.

    """
    _n_experts: int
    _eps: float
    _weights: List[float]
    _probs: List[float]

    def __init__(self, n_experts: int, eps: float):
        assert n_experts > 0
        assert eps > 1e-6
        assert eps < 0.5-1e-6

        self._n_experts = n_experts
        self._eps = eps
        self._weights = [1 for _ in range(self._n_experts)]
        # here we use _n_experts+1 because of the extra dimension to transform the
        # equality constraint into <=.
        self._probs = [1/(self._n_experts+1) for _ in range(self._n_experts)]

    def update_weights(self, expert_gains: List[float]) -> List[float]:
        """Update mwu probailitiesbased on a new input of the cost function.

        Parameters
        ----------
        expert_gains : list of float
            A list containing the new input from the cost function. Length must match instance dimension.

        Returns
        -------
        list of float
            Updated probabilies of each dimension.
        """
        assert len(expert_gains) == self._n_experts

        for i in range(self._n_experts):
            self._weights[i] *= (1 + self._eps * expert_gains[i])
        # we sum an extra 1 that is the extra expert weight
        sum_weights = sum(self._weights)+1
        for i in range(self._n_experts):
            self._probs[i] = self._weights[i] / sum_weights
        return copy.copy(self._probs)

    def get_probs(self) -> List[float]:
        """Getter for the current probabilities.

        Returns
        -------
        list of float
            Current probabilies of each dimension.
        """
        return copy.copy(self._probs)

    def get_n_experts(self) -> int:
        """Getter for the number of dimensions (experts) in this instance.

        Returns
        -------
        int
            The number of experts.
        """
        return self._n_experts



"""This module helps with generation of instances for our problem.
"""
from typing import List, Tuple, Callable
import random
from copy import deepcopy
from math import log, sqrt, floor


def _find_cap(target_delta: float, cost_dim: int) -> float:
    """Find capacity such that epsilon <= 0.5 (MWU guarantees) and
    delta < target_delta (so we dont waste many values estimating Z).

    Here we consider only integer values for the capacity, and
    use binary search to find the smallest valid capacity in
    logarithmic time.
    """
    # function to compute the best (lowest) epsilon
    e: Callable[[float], float] = lambda c: sqrt(log(cost_dim, 2)/c)
    # function to compute hardest/"most restrictive" (lowest) delta based on capacity
    delta: Callable[[float], float] = lambda c: (12 / c) * log((cost_dim+2)*c / log(cost_dim))

    # Binary search to find ideal capacity
    left: float = 12 * log((cost_dim+2)/log(cost_dim))
    right: float = 1e10
    while left < right:
        mid: float = floor((left + right) / 2)
        if (delta(mid) > target_delta or e(mid) > 0.5-1e-6):
            left = mid+1
        else:
            right = mid
    return left


def _get_random_values(n_instants: int, items_per_instant: int) -> List[List[float]]:
    """Generates the list of items values randomly.
    """
    return [[random.random() for _ in range(items_per_instant)]
            for _ in range(n_instants)]


def _get_random_costs(n_instants: int, items_per_instant: int, cost_dim: int) -> List[List[List[float]]]:
    """Generates the list of items costs randomly.
    """
    weights: List[List[List[float]]] = [[] for _ in range(n_instants)]
    for t in weights:
        for _ in range(items_per_instant):
            t.append([random.random() for _ in range(cost_dim)])
    return weights


def generate_valid_instance(target_delta: float, n_instants: int, cost_dim: int,
                            items_per_instant: int = 1) -> \
        Tuple[List[List[float]], List[List[List[float]]], float, float]:
    """Generates values for a problem instance that respect guarantees premises,
    thus, theoric guarantees should be valid (see Notes section below).

    Parameters
    ----------
    target_delta : float
        Capacity will be set with the constraint that delta <= `target_delta`.
    n_instants : int
        Number of instants to be generated.
    cost_dim : int
        Dimension of the cost vectors to be generated.
    items_per_instant : int
        Number of items that should be available in each instant.

    Returns
    -------
    values : list of list of float
        A list containing, for each instant, a list with that instant item's values.
    costs : list of list of list of float
        A list containing, for each instant, a list with that instant item's cost vectors.
    cap : float
        A random problem capacity.
    e : float
        The best theorical epsilon for the generated problem.

    Notes
    -----
    `target_delta` should be calibrated relative to `n_instants`. Setting a `target_delta`
    too low can cause the :math:`optimum\\_value\\_sum \\geq \\log d / \\epsilon^2` premise to be violated.
    Setting `target_delta` too high is not ideal since the algorithm chooses the items randomly in the first
    `target_delta` fraction of instants.

    The only premise that is guaranteed here is that :math:`cap \\geq \\log d / \\epsilon^2`. The
    :math:`optimum\\_value\\_sum \\geq \\log d / \\epsilon^2` premise is not asserted since it would
    require solving the instance to find `optimum_value_sum`, making this function slow. Setting
    a higher `target_delta` will increase the chance that this premise is valid.
    """
    assert items_per_instant > 0
    assert cost_dim > 0
    assert target_delta + 1e-6 < 1
    assert target_delta - 1e-6 > 0

    values: List[List[float]] = _get_random_values(n_instants, items_per_instant)
    costs: List[List[List[float]]] = _get_random_costs(n_instants, items_per_instant, cost_dim)

    cap = _find_cap(target_delta, cost_dim)
    e = sqrt(log(cost_dim)/cap)
    return values.copy(), deepcopy(costs), cap, e


def generate_random_instance(n_instants: int, cost_dim: int, items_per_instant: int = 1) -> \
        Tuple[List[List[float]], List[List[List[float]]], float, float]:
    """Generates random values, costs and capacity for a Packing Problem instance.
    Instances generated here may not respect guarantees constraints.

    Parameters
    ----------
    n_instants : int
        Number of instants to be generated.
    cost_dim : int
        Dimension of the cost vectors to be generated.
    items_per_instant : int
        Number of items that should be available in each instant.

    Returns
    -------
    values : list of list of float
        A list containing, for each instant, a list with that instant item's values.
    costs : list of list of list of float
        A list containing, for each instant, a list with that instant item's cost vectors.
    cap : float
        A random problem capacity.
    e : float
        The best theorical epsilon for the generated problem.
    """
    assert items_per_instant > 0
    assert cost_dim > 0

    values: List[List[float]] = _get_random_values(n_instants, items_per_instant)
    costs: List[List[List[float]]] = _get_random_costs(n_instants, items_per_instant, cost_dim)

    cap = random.random() * n_instants/2
    e = sqrt(log(cost_dim, 2)/cap)
    return values.copy(), deepcopy(costs), cap, e



from typing import Any, Union, List
from mip import Model, MAXIMIZE, CBC, BINARY, OptimizationStatus, xsum, maximize
from fast_online_packing.offline_solvers.base_solver import BaseSolver


class PythonMIPSolver(BaseSolver):
    """Python-MIP Solver, used to model and solve the offline Packing LP Problem,
    using a COIN-OR Branch and Cut (CBC) solver.

    Parameters
    ----------
    values
        Refer to the BaseSolver parameters.
    costs
        Refer to the BaseSolver parameters.
    capacity
        Refer to the BaseSolver parameters.
    max_seconds : int, default=30
        Maximum time spent searching for better solutions, in seconds.

    Attributes
    ----------
    optimum_value : float or int
        The optimum value found for the LP problem.
    packed_items : list of int
        A list containing the indexes of the items chosen in each instant.
    packed_weight_sum : list of float
        A list containing the optimal solution's total cost for each dimension.
    solver
        The instance of the Python-MIP solver.
    max_seconds : int
        Maximum time spent searching for better solutions, in seconds.
    status : {mip.OptimizationStatus.OPTIMAL, mip.OptimizationStatus.FEASIBLE, \
              mip.OptimizationStatus.INFEASIBLE}
        Result of the optimization procedure.

    Methods
    -------
    solve()
        Solves the LP problem.
    print_result()
        Inherited from BaseSolver.

    Notes
    -----
    This is a Mixed Integer Programming modeler and solver, being much more flexible than
    the Google Knapsack Solver, accepting all types of variantions of the problem: more than one
    item per instant, obligatory packing or flexible packing.

    Read more about the `Python-MIP Solver`_.

    .. _Python-MIP Solver: https://docs.python-mip.com/en/latest/index.html
    """
    solver: Any
    max_seconds: int
    status: int

    def __init__(self, values: List[List[Union[float, int]]],
                 costs: List[List[List[Union[float, int]]]],
                 capacity: Union[float, int],
                 max_seconds: int = 30):
        super().__init__(values, costs, capacity)
        self.max_seconds = max_seconds
        self._build_model()

    def _build_model(self) -> None:
        """Initializes the Python-MIP solver and builds a model for the Linear Programming
        problem, setting an objective function and constraints.

        Returns
        -------
        None
        """
        self.solver = Model(sense=MAXIMIZE, solver_name=CBC)
        m = self.solver
        m.verbose = 0
        # creating variables for all instants and options
        x = [[m.add_var(name=f"x[{t}][{opt}]", var_type=BINARY)
              for opt in range(self._options_per_instant)]
             for t in range(self._size)]
        # setting capacity constraint
        for dim in range(self._cost_dimension):
            m += xsum(self._costs[i][item][dim]*x[i][item]
                      for item in range(self._options_per_instant)
                      for i in range(self._size)) <= self._capacity  # type: ignore
        # setting constraint to only choose one item per instant
        for t in range(self._size):
            m += xsum(x[t][i] for i in range(self._options_per_instant)) <= 1  # type: ignore
        # setting objective function
        m.objective = maximize(xsum(self._values[i][item]*x[i][item]
                                    for item in range(self._options_per_instant)
                                    for i in range(self._size)))

    def solve(self):
        """Solves the optimization problem and sets `status`, `optimum_value`,
        `packed_items` and `packed_weight_sum`.

        Returns
        -------
        None
        """
        self.packed_items = [-1] * self._size
        self.status = self.solver.optimize(max_seconds=self.max_seconds)
        if self.status in [OptimizationStatus.OPTIMAL, OptimizationStatus.FEASIBLE]:
            self.optimum_value = self.solver.objective_value
            for idx, v in enumerate(self.solver.vars):
                if abs(v.x) > 1e-6:  # chosen options for an instant are the positive variables
                    chosen_idx = idx % self._options_per_instant
                    time_instant = idx // self._options_per_instant
                    self.packed_items[time_instant] = chosen_idx
            for t, packed_idx in enumerate(self.packed_items):
                if t >= 0:
                    for dim in range(self._cost_dimension):
                        self.packed_weight_sum[dim] += self._costs[t][packed_idx][dim]
        else:
            self.optimum_value = float("inf")



"""Helper module that provides usable math functions.
"""
from typing import List, Union


def dot_product(a: List[Union[int, float]], b: List[Union[int, float]]) -> Union[int, float]:
    """Compute dot product between two vectors.

    Parameters
    ----------
    a : list of (float or int)
        First vector.
    b : list of (float or int)
        Second vector.

    Returns
    -------
    int or float
        The result of the dot product.

    Raises
    ------
    Exception
        If `a` and `b` have different lengths.

    """
    if len(a) != len(b):
        raise Exception(f"The two vectors must have the same length: len(a)={len(a)} != len(b)={len(b)}.")
    sum: float = 0.0
    for i in range(len(a)):
        sum += a[i] * b[i]
    return sum



"""This module contains an algorithm to solve the Online Packing Problem.

Notes
-----
The algorithm implemented here to solve the online packing problem
is described in [1]_, on section 6, as Algorithm 6.1. It uses the
MWU (Multiplicative Weights Update) method, as described in
[2]_, and Integer/Linear prorgamming solvers. Hence the `Mwu-Max`
and the `Offline Solvers` modules.

References
----------
.. [1] Agrawal, Shipra & Devanur, Nikhil. (2014). Fast Algorithms for Online Stochastic \
Convex Programming. Proceedings of the Annual ACM-SIAM Symposium on Discrete Algorithms. \
2015. 10.1137/1.9781611973730.93.

.. [2] Arora, Sanjeev & Hazan, Elad & Kale, Satyen. (2012). The Multiplicative Weights \
Update Method: a Meta Algorithm and Applications. Theory of Computing [electronic only]. \
8. 10.4086/toc.2012.v008a006.
"""

from math import log, sqrt, ceil
from typing import List, Union, Type
import random
from fast_online_packing.mwu_max import MwuMax
from operator import itemgetter
from fast_online_packing import helper
from fast_online_packing.offline_solvers.base_solver import BaseSolver
from fast_online_packing.offline_solvers.python_mip_solver import PythonMIPSolver
from fast_online_packing.packing_problem import PackingProblem


class OnlineSolver:
    """Solves the Online Packing Problem.

    Parameters
    ----------
    cost_dimension : int
        Dimension of the cost vectors to be used.
    total_time : int
        Total number of instants.
    capacity : float
        Capacity restriction of the problem instance.
    e : float or None, default=None
        Epsilon parameter to be used in the algorithm. If none is given, \
            a theorical optimum epsilon is set.
    solver_cls : Type[BaseSolver]
        Offline solver to be used.

    Attributes
    ----------
    p : PackingProblem
        Packing Problem instance, that holds and enforces all problem-related aspects.
    e : float
        Epsilon parameter used in the algorithm.
    mwu : MwuMax
        MWU instance.
    z : Union[float, None]
        Algorithm calculated parameter Z.
    delta : float
        Decimal fraction of instants used to estimate Z.
    total_time : int
        Total number of instants in the algorithm.
    current_time : int
        Count of the instant the algorithm currently is. 1-indexed.
    optimum_value : float
        Sum of the values of the items packed in the optimal solution (solving offline).
    cost_dimension : int
        Property that gets the dimension of the problem instance's cost vectors.

    Methods
    -------
    print_params()
        Print parameters used in the algorithm.
    pack_one(available_values, available_costs)
        Choose one of the given items to pack on the current instant.
    compute_optimum()
        Solves the offline problem for the previously set instants, \
            in order to find the optimum solution.
    print_result()
        Print a results report.
    get_premises_min()
        Gets the minimum value of the capacity and optimum-value-sum so that \
            premises and theoric guarantees are valid.
    respect_premises()
        Informs if the algorithm is in agreement with the theoric premises.
    """

    p: PackingProblem
    e: float
    mwu: MwuMax
    z: Union[float, None]
    delta: float
    total_time: int
    current_time: int
    optimum_value: float

    # TODO: receive different solvers in this function
    def __init__(self, cost_dimension: int, total_time: int, capacity: float, e: Union[float, None] = None,
                 solver_cls: Type[BaseSolver] = PythonMIPSolver):
        self.z = None
        self.p = PackingProblem(capacity, cost_dimension)
        self._init_params(cost_dimension, total_time, capacity, e)
        self._solver_cls = solver_cls
        self.optimum_value = float("inf")

    def _init_params(self, cost_dimension: int, total_time: int, capacity: float,
                     e: Union[float, None]) -> None:
        """Initializes method parameters
        """
        if e is None:
            self.e = sqrt(log(cost_dimension, 2)/capacity)
        else:
            assert e + 1e-6 < 0.5
            assert e - 1e-6 > 0
            self.e = e
        self.z = None
        e_2 = self.e * self.e
        self.delta = 12 * e_2 * log((cost_dimension+2)/e_2) / log(cost_dimension)
        self.current_time = 1
        self._initial_phase_size = int(ceil(self.delta * total_time))
        self.total_time = total_time
        self.mwu = MwuMax(self.cost_dimension, self.e)

    @property
    def cost_dimension(self):
        return self.p.get_cost_dimension()

    def print_params(self):
        """Print algorithm parameters.
        """
        print(f"capacity = {self.p.get_capacity()}")
        print(f"cost dimension = {self.cost_dimension}")
        print(f"e = {self.e}")
        print(f"delta = {self.delta}")
        print(f"initial phase size = {self._initial_phase_size}")
        print(f"total time = {self.total_time}")

    def _compute_z(self) -> float:
        """Compute Z, by solving a scaled offline problem described in the paper's apendix.

        Returns
        -------
        float
            Z parameter
        """
        scaled_cap = (self.delta * self.p.get_capacity() +
                      sqrt(3 * self.delta * self.p.get_capacity() *
                           log((self.cost_dimension+2)/(self.e * self.e))))
        s = self._solver_cls(self.p.get_available_values(), self.p.get_available_costs(), scaled_cap)
        s.solve()
        return 2 * s.optimum_value / (self.delta * self.p.get_capacity())

    def _choose_index_to_pack(self, available_values: List[float],
                              available_costs: List[List[float]]) -> int:
        """Given available items for an instant, chooses which item to pack.

        Parameters
        ----------
        available_values : list of float
            Value of the items available on the current instant.
        available_costs : list of list of float
            Cost vectors of the items available on the current instant.

        Returns
        -------
        int
            Index of the item with the highest evaluated value.
        """
        if self.current_time <= self._initial_phase_size:
            return random.randint(-1, len(available_values)-1)
        else:
            if self.z is None:
                self.z = self._compute_z()
            evaluated_options = [available_values[i]
                                 - self.z * helper.dot_product(self.mwu.get_probs(), available_costs[i])
                                 for i in range(len(available_values))]
            # item that has the highest evaluated value from evaluated_options
            max_idx, max_value = max(enumerate(evaluated_options), key=itemgetter(1))
            # evaluate if its worth to get the item or not
            return max_idx if max_value > 0+1e-6 else -1

    def _compute_mwu_gains(self, cost: float) -> float:
        """Compute MWU cost(gains) function for a single dimension.

        Parameters
        ----------
        cost : float
            Cost of a single dimension of the cost vector.

        Returns
        -------
        float
            Gains of a single dimension
        """
        return cost - self.p.get_capacity()/self.total_time

    def pack_one(self, available_values: List[float], available_costs: List[List[float]]) -> int:
        """Receives the new instant's items and chooses one to pack.

        Parameters
        ----------
        available_values : list of float
            Value of the items available on the current instant.
        available_costs : list of list of float
            Cost vectors of the items available on the current instant.

        Returns
        -------
        int
            Index of the chosen item or -1 if none of the items were packed.
        """
        self.p.set_current_inputs(available_values, available_costs)
        chosen_idx = self._choose_index_to_pack(available_values, available_costs)
        if not self.p.item_fits(chosen_idx):
            chosen_idx = -1
        self.p.pack(chosen_idx)
        received_costs = available_costs[chosen_idx] if chosen_idx != -1 else [0]*self.cost_dimension
        mwu_gains = list(map(self._compute_mwu_gains, received_costs))
        self.mwu.update_weights(mwu_gains)
        self.current_time += 1
        return chosen_idx

    def compute_optimum(self) -> float:
        """Solve the problem offline to find out the optimum solution.

        Returns
        -------
        The optimum solution for the problem.
        """
        s = self._solver_cls(self.p.get_available_values(),
                             self.p.get_available_costs(), self.p.get_capacity())
        s.solve()
        self.optimum_value = s.optimum_value
        return s.optimum_value

    def print_result(self) -> None:
        """Print usefull information about the algorithm execution.
        """
        print(f"Opt: {self.optimum_value}")
        print(f"Alg: {self.p._packed_value_sum}")  # type: ignore
        print(f"Score Alg = {self.p._packed_value_sum/self.optimum_value :.3f} * Opt")  # type: ignore
        print(f"     min {{B, TOPT}} = {self.get_premises_min():.3f}")
        print(f"     B = {self.p.get_capacity()}")
        print(f"     TOPT = {self.optimum_value}")

    def get_premises_min(self) -> float:
        """Get the minimum value of capacity and `optimum_value` for the
        premises to be valid.

        Returns
        -------
        float
            Minimum value of capacity and `optimum_value` for premises to be valid.
        """
        return log(self.p.get_cost_dimension()) / (self.e*self.e)

    def respect_premises(self) -> bool:
        """Informs if both the algorithm premises (`optimum_value` and capacity) were fulfilled.

        Returns
        -------
        bool
            True if both premises were fulfilled, False otherwise.
        """
        if min(self.optimum_value, self.p.get_capacity()) + 1e-6 < self.get_premises_min():
            return False
        else:
            return True



from ortools.algorithms import pywrapknapsack_solver
from typing import List, Any, Tuple, Union
from fast_online_packing.offline_solvers.base_solver import BaseSolver


class GoogleKnapsackSolver(BaseSolver):
    """Google OR-Tools solver for the multi-dimensional Knapsack
    problem using integer values.

    Parameters
    ----------
    values
        Refer to the BaseSolver parameters.
    costs
        Refer to the BaseSolver parameters.
    capacity
        Refer to the BaseSolver parameters.
    decimal_places : int
        Decimal precision used to scale floats in values and costs into integers.

    Attributes
    ----------
    optimum_value : float or int
        The optimum value found for the LP problem.
    packed_items : list of int
        A list containing the indexes of the items chosen in each instant.
    packed_weight_sum : list of float
        A list containing the optimal solution's total cost for each dimension.
    solver
        The instance of the used OR-Tools `pywrapknapsack_solver.KnapsackSolver` solver.
    adapted_values : list of int
        `values` scaled into an integer using `decimal_places` decimal places as precision.
    adapted_costs : list of list of int
        `costs` scaled into an integer using `decimal_places` decimal places as precision.
    adapted_capacity : list of int
        A list with, for every cost dimension, the same `capacity` scaled into an integer using
        `decimal_places` decimal places as precision.
    decimal_places : int
        Decimal precision used to scale floats in `values` and `costs` into integers.

    Methods
    -------
    solve()
        Solves the LP problem.
    print_result()
        Inherited from BaseSolver.
    validate_instance(values, costs)
        Check if an offline packing LP instance is valid for this solver.

    Notes
    -----
    The google solver used here does not accept floating point numbers, as it considers the
    problem as a discrete one. Thus, we multiply every value, cost and capacity by
    :math:`10^n` where :math:`n` is `decimal_places`.

    Also, this solver only accepts one item options per instant.

    *Validity Conditions:*
    Also, there are variantions of the packing problem. Some allow the algorithm
    to do nothing, not packing any item, in an instant. Some compel the algorithm
    to make a choice in every instant. The google solver only accepts instances
    where it is not required to pack an item in every instant. This solver also only
    accepts instances where there is only one item available per instant.

    Read more about the `Google OR-Tools Knapsack Solver`_.

    .. _Google OR-Tools Knapsack Solver: https://developers.google.com/optimization/bin/knapsack
    """
    solver: Any
    adapted_values: List[int]
    adapted_costs: List[List[int]]
    adapted_capacity: List[int]
    decimal_places: int

    def __init__(self, values: List[List[Union[float, int]]],
                 costs: List[List[List[Union[float, int]]]],
                 capacity: float,
                 decimal_places: int = 6):
        GoogleKnapsackSolver.validate_instance(values, costs)
        super().__init__(values, costs, capacity)
        self.decimal_places = decimal_places
        self.adapted_values, self.adapted_costs, self.adapted_capacity = self._adapter()
        self.solver = pywrapknapsack_solver.KnapsackSolver(pywrapknapsack_solver.KnapsackSolver
                                                           .KNAPSACK_MULTIDIMENSION_BRANCH_AND_BOUND_SOLVER,
                                                           'KnapsackExample')
        self.solver.Init(self.adapted_values, self.adapted_costs, self.adapted_capacity)

    @staticmethod
    def validate_instance(values: List[List[Union[float, int]]],
                          costs: List[List[List[Union[float, int]]]]) -> None:
        """Check if a given instance is valid for this solver. Refer to the
        *Notes* to check validity conditions.

        Parameters
        ----------
        values
            Values for the instance. For format, refer to BaseSolver parameters.
        costs
            Costs for the instance. For format, refer to BaseSolver parameters.

        Returns
        -------
        None

        Raises
        ------
        Exception
            If problem instance is not valid.
        """
        for value_options, cost_options in zip(values, costs):
            if(len(value_options) > 1 or len(cost_options) > 1):
                raise Exception("GoogleKnapspackSolver only works with one item per instant.")

    def _adapter(self) -> Tuple[List[int], List[List[int]], List[int]]:
        """Adapts instance values from floats into integers.

        Returns
        -------
        values : list of int
            List containing the adapted values of each instant (only one item per instant in this solver).
        costs : list of list of int
            List containing, for each instant, the item's adapted costs.
        capacity : list of int
            List containing adapted capacity, the same for each dimension.

        """
        factor = pow(10, self.decimal_places)
        capacity = [int(self._capacity * factor)] * self._cost_dimension
        values = [int(t[0]*factor) for t in self._values]
        costs = [[int(self._costs[t][0][dim]*factor)
                  for t in range(self._size)]
                 for dim in range(self._cost_dimension)]
        return values, costs, capacity

    def solve(self):
        """Try to solve the LP problem and sets `optimum_value`, `packed_items`
        and `packed_weight_sum`.

        Returns
        -------
        None
        """
        factor = pow(10, self.decimal_places)
        self.optimum_value: float = self.solver.Solve() / factor
        self.packed_items = [-1] * self._size

        for i in range(len(self.adapted_values)):
            # if item was packed it was first option of the instant, meaning index 0
            if self.solver.BestSolutionContains(i):
                self.packed_items[i] = 0
                for dim in range(self._cost_dimension):
                    self.packed_weight_sum[dim] += self._costs[i][0][dim]
